# -*- coding: utf-8 -*-
import os
import time
import re
import xml.dom.minidom
from operator import itemgetter, attrgetter

def new_vcs(project_name):
    """Factory Method that creates the right type (SVN or GIT) of VCS for the current dir"""
    def is_curdir_svn_workdir():
	return '.svn' in os.listdir('.')

    if is_curdir_svn_workdir():
        return SVN(project_name)
    else:
        return GIT(project_name)

class VCS(object):
    """Class representing a Version Control System.

    All VCS subclasses will have to implement ALL the abstract methods below.
    The scons build script will expect a VCS supporting all the operation below.
    """

    def all_tags(self):
        """Return a list of all the tag names"""
        raise NotImplementedError

    def latest_tag(self):
        """Return the name of the latest tag"""
        raise NotImplementedError

    def in_trunk(self):
        """Return True if the working directory is on trunk/master"""
        raise NotImplementedError

    def checkout_tag_command(self, tag, co_dir):
        """Return a command (as string) that will check out the given tag in the given directory"""
        raise NotImplementedError

    def checkout_trunk_command(self, co_dir):
        """Return a command (as sting) that will check out trunk/master in the given directory"""
        raise NotImplementedError

    def current_tag_name(self):
        """Return the name of the tag the working directory is on or None for trunk/master"""
        raise NotImplementedError

def run_command(command, arguments):
    return os.popen("%s %s" % (command, arguments)).read().strip()

def last_item(l):
    return l[-1] if l else None

class SVN(VCS):
    def __init__(self, project_name):
        self.project_name = project_name
        self.svn = run_command('/usr/bin/which', 'svn')
        assert self.svn, "It seems that SVN isn't installed"
        self.url = self.get_working_dir_url()
        self.trunk_url = self.get_trunk_url()
        self.tags_url = self.get_tags_url()

    def all_tags(self):
        return [ tag[1] for tag in self.list_tags() ]

    def latest_tag(self):
        last = last_item(self.list_tags())
        return last[1] if last else None

    def in_trunk(self):
        return self.url.startswith(self.trunk_url)

    def checkout_tag_command(self, tag, co_dir):
        return self.checkout_url_command(self.tag_url(tag), co_dir)

    def checkout_trunk_command(self, co_dir):
        return self.checkout_url_command(self.trunk_url, co_dir)

    def current_tag_name(self):
        parts = self.url.split('/')
        if "tags" not in parts:
            return None
        return parts[parts.index("tags")+1]

    # private

    def tag_url(self, tag):
        return "%s/%s" % (self.tags_url, tag)

    def checkout_url_command(self, url, co_dir):
        return "%s co %s %s" % (self.svn, url, co_dir)

    def get_working_dir_url(self):
        dom = xml.dom.minidom.parseString(
            run_command(self.svn, 'info --xml'))
        e = dom.getElementsByTagName('entry')[0]
        url = dom.getElementsByTagName('url')[0].childNodes[0].data
        dom.unlink()
        return url
    
    def get_base_url(self):
	"""Return the parent directory of tags, trunk or branches"""
        parts = self.url.split('/')
        rparts = parts[:]
        rparts.reverse()

        # work back from the tail until we find the 'trunk' item
        for bit in rparts:
            parts.pop()
            if bit in ("trunk", "tags", "branches"):
                break 

        return "/".join(parts) if parts else None

    def get_trunk_url(self):
	return "/".join([self.get_base_url(), "trunk"])

    def get_tags_url(self):
	return "/".join([self.get_base_url(), "tags"])

    def list_tags(self):
        dom = xml.dom.minidom.parseString(
            run_command(self.svn, 'list %s --xml' % self.tags_url))
        entries = dom.getElementsByTagName('entry')
        out=[]
        for entry in entries:
            name = entry.getElementsByTagName('name')[0].childNodes[0].data
            author = entry.getElementsByTagName('author')[0].childNodes[0].data
            datestring = entry.getElementsByTagName('date')[0].childNodes[0].data
            date = time.strptime(datestring.split('.')[0],"%Y-%m-%dT%H:%M:%S")
            revision = int(entry.getElementsByTagName('commit')[0].getAttribute('revision'))
            out.append( (revision, name, author, date) )
        out.sort(key=itemgetter(3))
        return out
        
    def get_tag_date(self, tag):
        taginfo = self.list_tags()
        
class GIT(VCS):
    TrunkBranchName = "master"

    def __init__(self, project_name):
        self.project_name = project_name
        self.git = run_command('/usr/bin/which', 'git')
        assert self.git, "It seems that GIT isn't installed"

    def all_tags(self):
        return run_command(self.git, 'tag').split()

    def latest_tag(self):
        # We take all the tags starting with release, extract the numbers from them
        release_tags = [ tag for tag in self.all_tags() if tag.upper().startswith('RELEASE') ]
        # RELEASE_1, RELEASE_1_2, RELEASE_1_2_2) becomes
        # [  [[1], 'RELEASE_1'], [[1,2], 'RELEASE_1_2'], [[1,2,2],'RELEASE_1_2_2'] ] 
        versioned_release_tags = zip(
            [[int(d) for d in re.split('\D+', tag) if d] for tag in release_tags],
            release_tags
        )
        last = last_item(sorted(versioned_release_tags))
        return last[1] if last else None

    def in_trunk(self):
        return not os.path.exists('.tag')
        #return self.in_branch(GIT.TrunkBranchName)

    def checkout_tag_command(self, tag, co_dir):
        # This implementation is far from ideal, but it should be temporary
        dir = "%s/%s" % (co_dir, self.project_name)
        svn_export_imitation = "%s archive --format=tar --prefix=%s/ %s | tar xf -" % (self.git, dir, tag)
        copy_build_dir = "cp -R build/ %s/" % dir
        tag_dir = tag if tag != GIT.TrunkBranchName else "trunk"
        add_tag_info = 'echo "%s" > %s/.tag' % (tag_dir,dir)
        return ";".join((svn_export_imitation, copy_build_dir, add_tag_info))

    def checkout_trunk_command(self, co_dir):
        return self.checkout_tag_command(GIT.TrunkBranchName, co_dir)

    def current_tag_name(self):
        return run_command('cat', '.tag')
        #describe_out = run_command(self.git, 'describe --all')
        #if describe_out.startswith("tags/"):
        #    return describe_out[len("tags/"):]

    # private 

    def in_branch(self, branch_name):
        return "heads/" + branch_name == run_command(self.git, 'describe --all')

    def in_tag(self, tag_name):
        return self.current_tag_name() == tag_name

